<?php

namespace Drupal\commerce_forteecom;

use Drupal;
use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_payment\Entity\PaymentInterface;
use Drupal\commerce_payment\Entity\PaymentMethodInterface;
use Drupal\commerce_payment\Exception\PaymentGatewayException;

class ForteEcomHelper
{
  const HOST = "https://ecom-merchant-api.fortebank.com";
  const HOST_DEMO = "https://qa-ecom-merchant-api.fortebank.com";

  const ENDPOINT_INIT = "/init";
  const ENDPOINT_STATUS = "/status";
  const ENDPOINT_COMPLETE = "/complete";
  const ENDPOINT_PURCHASE = "/purchase";
  const ENDPOINT_PURCHASE_TOKEN = "/token/purchase";
  const ENDPOINT_REFUND = "/refund";
  const ENDPOINT_REVERSE = "/reverse";

  const ENDPOINT_RESPONSE_CODE = "https://adminapi.forte.kz/codes";

  const STATUS_INIT = 'CREATED';
  const STATUS_LOCK = 'ON_LOCK';
  const STATUS_PROCESS = 'ON_PAYMENT';
  const STATUS_SUCCESS = 'APPROVED';
  const STATUS_DECLINE = 'DECLINED';
  const STATUS_ON_HOLD = 'PREAUTH_APPROVED';
  const STATUS_REVERSE = 'REVERSED';
  const STATUS_REFUND = 'REFUNDED';
  const STATUS_EXPIRE = 'EXPIRED';
  const STATUS_CANCEL = 'CANCELED';
  const STATUS_ERROR = 'ERROR';

  const AVAILABLE_CURRENCIES = [
    'KZT' => [
      'id'   => 398,
      'code' => 'KZT',
      'name' => 'Kazakhstani tenge'
    ],
  ];

  /**
   * Initialize ForteEcom payment
   *
   * @param string $_forteecom_merchant_id
   * @param bool   $pre_auth
   * @param float  $amount
   * @param int    $remote_id   Order id on merchant side
   * @param string $description Order description on merchant side
   * @param string $currency_code
   *
   * @return array
   */
  public static function init_payment($_forteecom_merchant_id, $pre_auth, $amount, $remote_id, $description, $currency_code)
  {
    $currency_id = self::get_currency_id($currency_code);
    $data        = [
      'merchantId' => $_forteecom_merchant_id,
      'preAuth'    => $pre_auth,
      'source'     => [
        'jsonClass'   => 'TransactionSourcePurchase',
        'amount'      => (float)$amount,
        'currency'    => $currency_id,
        'remoteId'    => $remote_id,
        'description' => $description,
      ]
    ];

    $request_body = json_encode($data);

    $request_result = self::send_request(self::ENDPOINT_INIT, $request_body);

    $_forteecom_order_id   = null;
    $_forteecom_session_id = null;
    if (empty($request_result['error'])) {
      if ($request_result['response']->responseCode == 0) {
        $_forteecom_order_id   = $request_result['response']->data->orderId;
        $_forteecom_session_id = $request_result['response']->data->sessionId;
      } else {
        $request_result['error'] = $request_result['response']->responseMessage;
      }
    }

    if (!empty($request_result['error'])) {
      throw new PaymentGatewayException('Could not initialize order in ForteEcom');
    }

    return [
      'error_message'         => $request_result['error'],
      '_forteecom_order_id'   => $_forteecom_order_id,
      '_forteecom_session_id' => $_forteecom_session_id,
    ];
  }

  /**
   * Get currency id from currency code (KZT => 398)
   *
   * @param $currency_code
   *
   * @return mixed
   */
  public static function get_currency_id($currency_code)
  {
    return self::AVAILABLE_CURRENCIES[$currency_code]['id'];
  }

  /**
   * @param string $endpoint
   * @param        $body
   *
   * @return array
   */
  public static function send_request($endpoint, $body)
  {
    $logger = Drupal::logger('commerce_forteecom');

    $curl = curl_init();
    $url = self::get_url($endpoint);
    curl_setopt_array($curl, [
      CURLOPT_URL            => $url,
      CURLOPT_RETURNTRANSFER => true,
      CURLOPT_FOLLOWLOCATION => true,
      CURLOPT_SSL_VERIFYHOST => 0,
      CURLOPT_SSL_VERIFYPEER => 0,
      CURLOPT_ENCODING       => '',
      CURLOPT_MAXREDIRS      => 10,
      CURLOPT_TIMEOUT        => 30,
      CURLOPT_HTTP_VERSION   => CURL_HTTP_VERSION_1_1,
      CURLOPT_CUSTOMREQUEST  => 'POST',
      CURLOPT_POSTFIELDS     => $body,
      CURLOPT_HTTPHEADER     => [
        'Content-Type: application/json',
      ]
    ]);

    $response = json_decode(curl_exec($curl));
    $error    = curl_error($curl);
    curl_close($curl);

    if (empty($error) && !property_exists($response, 'responseCode')) {
      $error = property_exists($response, 'moreInfo') ? $response->moreInfo : null;
    }

    $log_data = [
      'url'      => $url,
      'request'  => $body,
      'response' => $response,
      'error'    => $error,
    ];
    $logger->info(print_r($log_data, true));

    return [
      'response' => $response,
      'error'    => $error
    ];
  }

  /**
   * @param string                 $_forteecom_merchant_id
   * @param string                 $_forteecom_order_id
   * @param string                 $_forteecom_session_id
   * @param float                  $amount
   * @param int                    $remote_id
   * @param string                 $description
   * @param string                 $currency_code
   * @param PaymentMethodInterface $payment_method
   *
   * @return array
   */
  public static function purchase_by_token($_forteecom_merchant_id, $_forteecom_order_id, $_forteecom_session_id, $amount, $remote_id, $description, $currency_code, $payment_method)
  {
    $currency_id = self::get_currency_id($currency_code);
    $data        = array(
      'merchantId' => $_forteecom_merchant_id,
      'orderId'    => $_forteecom_order_id,
      'sessionId'  => $_forteecom_session_id,
      'ip'         => $_SERVER['REMOTE_ADDR'],
      'token'      => $payment_method->getRemoteId(),
      'source'     => array(
        'jsonClass'   => "TransactionSourcePurchase",
        'amount'      => (float)$amount,
        'currency'    => $currency_id,
        'remoteId'    => $remote_id,
        'description' => $description
      ),
    );

    $request_body   = json_encode($data);
    $request_result = self::send_request(self::ENDPOINT_PURCHASE_TOKEN, $request_body);

    $_forteecom_order_status = null;
    $_forteecom_order_data   = null;
    if (empty($request_result['error'])) {
      $_forteecom_order_status = $request_result['response']->responseMessage;
      if ($request_result['response']->responseCode != 0) {
        $_forteecom_order_data   = $request_result['response']->data->purchaseResponseCode;
        $request_result['error'] = self::get_response_message($request_result['response']->data->purchaseResponseCode);
      } else {
        $_forteecom_order_data = $request_result['response']->data->result;
      }
    }

    return array(
      'error_message'           => $request_result['error'],
      '_forteecom_order_status' => $_forteecom_order_status,
      '_forteecom_order_data'   => $_forteecom_order_data,
    );

  }

  /**
   * Get localized message for error
   *
   * @param string $code
   *
   * @return string
   */
  public static function get_response_message($code)
  {
    $store_locale = Drupal::languageManager()->getCurrentLanguage()->getId();
    switch ($store_locale) {
      case 'ru':
        $lang = '_ru';
        break;
      case 'kk':
        $lang = '_kk';
        break;
      default:
        $lang = '';
        break;
    }

    $url = self::ENDPOINT_RESPONSE_CODE . '?' . http_build_query(['name' => $code]);

    $curl = curl_init();
    curl_setopt_array($curl, array(
      CURLOPT_URL            => $url,
      CURLOPT_RETURNTRANSFER => true,
      CURLOPT_FOLLOWLOCATION => true,
      CURLOPT_SSL_VERIFYHOST => 0,
      CURLOPT_SSL_VERIFYPEER => 0,
      CURLOPT_ENCODING       => '',
      CURLOPT_MAXREDIRS      => 10,
      CURLOPT_TIMEOUT        => 30,
      CURLOPT_HTTP_VERSION   => CURL_HTTP_VERSION_1_1
    ));

    $response = json_decode(curl_exec($curl), true);
    $error    = curl_error($curl);
    curl_close($curl);

    return $response[0]['message' . $lang];
  }

  /**
   * Complete ForteEcom payment status
   *
   * @param string $_forteecom_merchant_id
   * @param string $_forteecom_order_id
   * @param string $_forteecom_session_id
   *
   * @return array
   */
  public static function get_payment_status($_forteecom_merchant_id, $_forteecom_order_id, $_forteecom_session_id)
  {
    $data = [
      'merchantId' => $_forteecom_merchant_id,
      'orderId'    => $_forteecom_order_id,
      'sessionId'  => $_forteecom_session_id,
    ];

    $request_body = json_encode($data);

    $request_result = self::send_request(self::ENDPOINT_STATUS, $request_body);

    $_forteecom_order_status = null;
    $_forteecom_add_info     = null;
    if (empty($error_message)) {
      $_forteecom_order_status = $request_result['response']->data->orderStatus;
      $_forteecom_add_info     = $request_result['response']->data->addInfo;
    }

    return [
      'error_message'           => $request_result['error'],
      '_forteecom_order_status' => $_forteecom_order_status,
      '_forteecom_add_info'     => $_forteecom_add_info,
    ];
  }

  /**
   * Complete ForteEcom "pre-auth"-ed payment
   *
   * @param string $_forteecom_merchant_id
   * @param string $_forteecom_order_id
   * @param string $_forteecom_session_id
   *
   * @return array
   */
  public static function complete_payment($_forteecom_merchant_id, $_forteecom_order_id, $_forteecom_session_id)
  {
    $data = [
      'merchantId' => $_forteecom_merchant_id,
      'orderId'    => $_forteecom_order_id,
      'sessionId'  => $_forteecom_session_id,
    ];

    $request_body = json_encode($data);

    $request_result = self::send_request(self::ENDPOINT_COMPLETE, $request_body);

    $_forteecom_approval_code = null;
    if (empty($request_result['error'])) {
      $_forteecom_approval_code = $request_result['response']->data->result->approvalCode;
    }

    return [
      'error_message'            => $request_result['error'],
      '_forteecom_approval_code' => $_forteecom_approval_code,
    ];
  }

  public static function refund_payment($_forteecom_merchant_id, $_forteecom_order_id, $_forteecom_session_id, $amount, $currency_code)
  {
    $currency_id = self::get_currency_id($currency_code);
    $data        = [
      'orderIdent' => [
        'merchantId'  => $_forteecom_merchant_id,
          'orderId'   => $_forteecom_order_id,
          'sessionId' => $_forteecom_session_id,
        ],
      'transaction' => [
        'jsonClass' => "TransactionSourceRefund",
        'amount'    => (float)$amount,
        'currency'  => $currency_id
      ],
    ];

    $request_body = json_encode($data);

    $request_result = self::send_request(self::ENDPOINT_REFUND, $request_body);

    $_forteecom_message = null;
    if (empty($request_result['error']) && property_exists($request_result['response'], 'message') && $request_result['response']->message == 'OK') {
      $_forteecom_message = $request_result['response']->message;
    } else {
      $request_result['error'] = 'Error occured on refund';
    }

    return [
      'error_message'      => $request_result['error'],
      '_forteecom_message' => $_forteecom_message,
    ];
  }

  /**
   * Update order and payment data base on ForteEcom status
   *
   * @param OrderInterface   $order
   * @param PaymentInterface $payment
   * @param string           $new_status
   *
   * @throws Drupal\Core\Entity\EntityStorageException
   */
  public static function update_payment_status(&$order, &$payment, $new_status)
  {
    if ($payment->getRemoteState() == $new_status) {
      return;
    }
    $payment->setRemoteState($new_status);
    $message = null;

    switch ($new_status) {
      case self::STATUS_SUCCESS:
        $payment->setState('completed');
        break;
      case self::STATUS_DECLINE:
        $payment->setState('failed');
        self::unset_order_forteecom_data($order);
        break;
      case self::STATUS_EXPIRE:
        $payment->setState('expired');
        self::unset_order_forteecom_data($order);
        break;
      case self::STATUS_REVERSE:
        $payment->setState('reversed');
        break;
      case self::STATUS_REFUND:
        $payment->setState('refunded');
        break;
      case self::STATUS_CANCEL:
        $payment->setState('cancelled');
        break;
      case self::STATUS_ON_HOLD:
        $payment->setAuthorizedTime(time());
        $payment->setState('authorization');
        break;
    }
    $order->save();
    $payment->save();
  }

  /**
   * Delete order's ForteEcom data
   *
   * @param OrderInterface $order
   */
  public static function unset_order_forteecom_data(&$order)
  {
    $order->setData('commerce_forteecom', null);
  }

  /**
   * Check currency requirements
   *
   * @param $currency_code
   */
  public static function check_currency($currency_code)
  {
    if (!array_key_exists($currency_code, self::AVAILABLE_CURRENCIES)) {
      $message = 'Currently ForteEcom is enabled only for this currencies: ' . self::get_available_currency_names();
      Drupal::messenger()->addError($message);
      throw new PaymentGatewayException($message);
    }
  }

  /**
   * Returns list of supported currencies
   *
   * @return string
   */
  public static function get_available_currency_names()
  {
    $currencies = [];
    foreach (self::AVAILABLE_CURRENCIES as $currency) {
      $currencies[] = $currency['name'];
    }
    return implode(', ', $currencies);
  }

  /**
   * Add/update order's ForteEcom data
   *
   * @param OrderInterface $order
   * @param string         $_forteecom_order_id
   * @param string         $_forteecom_order_status
   * @param string         $_forteecom_session_id
   * @param string         $_forteecom_merchant_id
   */
  public static function update_order_forteecom_data(&$order, $_forteecom_order_id, $_forteecom_order_status, $_forteecom_session_id, $_forteecom_merchant_id)
  {
    $order->setData('commerce_forteecom', [
      '_forteecom_order_id'     => $_forteecom_order_id,
      '_forteecom_order_status' => $_forteecom_order_status,
      '_forteecom_session_id'   => $_forteecom_session_id,
      '_forteecom_merchant_id'  => $_forteecom_merchant_id,
    ]);
  }

  /**
   * @param PaymentInterface $payment
   *
   */
  public static function process_payment_status($payment)
  {
    $messenger = Drupal::messenger();

    switch ($payment->getRemoteState()) {
      case self::STATUS_SUCCESS:
        $messenger->addMessage('Payment is completed.');
        break;
      case self::STATUS_DECLINE:
        throw new PaymentGatewayException('Payment was not successful.');
        break;
      case self::STATUS_EXPIRE:
        throw new PaymentGatewayException('Payment is expired.');
        break;
      case self::STATUS_ON_HOLD:
        $messenger->addMessage('Payment is on-hold.');
        break;
      default:
        $messenger->addWarning('Payment is on-hold.');
        break;
    }
  }

  public static function get_url($endpoint)
  {
    $host = (FORTEECOM_DEMO_ENABLED) ? self::HOST_DEMO : self::HOST;
    return $host . '/v1/ecom' . $endpoint;
  }


}
